import logging
import sys
import json
from datetime import date
from pathlib import Path
from typing import AnyStr, Tuple


class Logger:
    LEVEL_DEFAULT = logging.INFO

    def __init__(self, logger_name, propagate=False):
        self.name = logger_name
        self.__logger_obj = None
        self.propagate = propagate
        self.file_handler = None

    @property
    def logger_obj(self) -> logging.Logger:
        if not self.__logger_obj:
            if self.logger_exists(self.name):
                self.__logger_obj = logging.getLogger(self.name)
            else:
                self.__logger_obj = logging.getLogger(self.name)
                self.__logger_obj.setLevel(self.LEVEL_DEFAULT)
                self.__logger_obj.propagate = self.propagate

                formatter = logging.Formatter(
                    "[%(asctime)s:%(module)s:%(funcName)s:"
                    "%(lineno)s:%(levelname)s] %(message)s"
                )

                stream_handler = logging.StreamHandler(sys.stderr)
                stream_handler.setFormatter(formatter)
                self.__logger_obj.addHandler(stream_handler)

        return self.__logger_obj

    @staticmethod
    def logger_exists(name: AnyStr) -> bool:
        return name in logging.Logger.manager.loggerDict.keys()

    def set_level(self, level):
        self.logger_obj.setLevel(level)

    def debug(self, msg: AnyStr, *args, **kwargs):
        self.logger_obj.debug(msg, *args, **kwargs)

    def info(self, msg: AnyStr, *args, **kwargs):
        self.logger_obj.info(msg, *args, **kwargs)

    def warning(self, msg: AnyStr, *args, **kwargs):
        self.logger_obj.warning(msg, *args, **kwargs)

    def error(self, msg: AnyStr, *args, **kwargs):
        self.logger_obj.error(msg, *args, **kwargs)

    def critical(self, msg: AnyStr, *args, **kwargs):
        self.logger_obj.critical(msg, *args, **kwargs)

    def log(self, level, msg: AnyStr, *args, **kwargs):
        self.logger_obj.log(level, msg, *args, **kwargs)

    def exception(self, msg: AnyStr, *args, **kwargs):
        self.logger_obj.exception(msg, *args, **kwargs)

    def log_to_file(self, level=logging.DEBUG):
        log_file_path: Path = self.get_log_file_path()

        try:
            if not log_file_path.parent.exists():
                log_file_path.parent.mkdir()
        except (OSError, IOError) as why:
            raise why

        self.file_handler = logging.FileHandler(log_file_path, mode="a")
        self.file_handler.setLevel(level)

        formatter = logging.Formatter(
            "[%(asctime)s:%(module)s:%(funcName)s:"
            "%(lineno)s:%(levelname)s] %(message)s"
        )

        self.file_handler.setFormatter(formatter)
        self.logger_obj.addHandler(self.file_handler)

    def stop_logging_to_file(self):
        if self.file_handler:
            self.logger_obj.removeHandler(self.file_handler)
            self.file_handler.close()
            self.file_handler = None

    def get_log_file_path(self) -> Path:
        user_path = Path("~").expanduser()
        module_dir = Path(__file__).parents[0]
        config_location = module_dir / "cfg/config.json"
        config = {}
        with open(config_location) as fp:
            config = json.load(fp)
        final_dir = user_path / f".{config.get('cfg_dir_name')}"
        today = date.today()
        date_string = today.strftime("%d-%m-%Y")
        log_file_path = final_dir / f"{self.logger_obj.name}_{date_string}.log"

        return log_file_path


def init_logger(
    base_name: AnyStr, log_to_file: bool = False
) -> Tuple[logging.Logger, logging.Logger]:
    """
    Initialize 'base_name' logging object and add a STDOUT handler to output to
    console, terminal, etc.

    Args:
        base_name (str): Base name for the logger. Will be used to create the logger name.
        log_to_file (bool): If True, the logger will log to a file.

    Returns:
        tuple: Tuple containing: logger, logger_gui (child of logger and used for GUI messages)
    """
    logger = Logger(f"{base_name}log")
    if log_to_file:
        logger.log_to_file()

    logger_gui = Logger(f"{base_name}log.gui", propagate=True)
    logger_gui.set_level(logging.INFO)

    return logger, logger_gui


logger, logger_gui = init_logger("vfxnaming")


if __name__ == "__main__":
    pass



class ParsingError(BaseException):
    """Raise when parsing couldn't be completed"""


class SolvingError(BaseException):
    """Raise when solving couldn't be completed"""


class TokenError(BaseException):
    """Raise when Token errors are detected."""


class RuleError(BaseException):
    """Raise when Rule errors are detected."""


class RepoError(BaseException):
    """Raise when Repo errors are detected."""



import copy
import json
from pathlib import Path
from typing import AnyStr, Dict, Union

from vfxnaming.serialize import Serializable
from vfxnaming.logger import logger
from vfxnaming.error import TokenError


_tokens = dict()


class Token(Serializable):
    def __init__(self, name: AnyStr, nice_name: AnyStr = ""):
        """Tokens are the meaningful parts of a naming rule. A token can be required,
        meaning fully typed by the user, or can have a set of default options preconfigured.
        If options are present, then one of them is the default one.
        Each option follows a {full_name:abbreviation} schema, so that names can be short
        but meaning can be recovered easily.

        Args:
            name (str): Name that best describes the Token, this will be used as a way
            to invoke the Token object.

            nice_name(str, optional): A more human readable name for the Token if needed for UI.
        """
        super(Token, self).__init__()
        self._name: AnyStr = name
        self._nice_name = name
        if len(nice_name):
            self._nice_name = nice_name
        self._default = None
        self._options: Dict = {}
        self._fallback = ""

    def add_option(self, fullname: AnyStr, abbreviation: AnyStr) -> bool:
        """Add an option pair to this Token.

        Args:
            fullname (str): Full name of the option
            abbreviation (str): Abbreviation to be used when building the name.

        Returns:
            [bool]: True if successful. False otherwise.
        """
        if fullname not in self._options.keys():
            self._options[fullname] = abbreviation
            if len(self._options) == 1:
                self._default = fullname
            return True
        logger.debug(
            f"Option '{fullname}':'{self._options.get(fullname)}' already exists in Token '{self.name}'. "
            "Use update_option() instead."
        )
        return False

    def update_option(self, fullname: AnyStr, abbreviation: AnyStr) -> bool:
        """Update an option pair on this Token.

        Args:
            fullname (str): Full name of the option
            abbreviation (str): Abbreviation to be used when building the name.

        Returns:
            [bool]: True if successful. False otherwise.
        """
        if fullname in self._options.keys():
            self._options[fullname] = abbreviation
            return True
        logger.debug(
            f"Option '{fullname}':'{self._options.get(fullname)}' doesn't exist in Token '{self.name}'. "
            "Use add_option() instead."
        )
        return False

    def remove_option(self, fullname: AnyStr) -> bool:
        """Remove an option on this Token.

        Args:
            key (str): Full name of the option

        Returns:
            [bool]: True if successful. False otherwise.
        """
        if fullname in self._options.keys():
            del self._options[fullname]
            return True
        logger.debug(
            f"Option '{fullname}':'{self._options.get(fullname)}' doesn't exist in Token '{self.name}'"
        )
        return False

    def clear_options(self):
        """Clears all the options for this token."""
        self._default = None
        self._options = {}

    def has_option_fullname(self, fullname: AnyStr) -> bool:
        """Looks for given option full name in the options.

        Args:
            fullname (str): Full name of the option

        Returns:
            [bool]: True if found. False otherwise.
        """
        if fullname in self._options.keys():
            return True
        return False

    def has_option_abbreviation(self, abbreviation: AnyStr) -> bool:
        """Looks for given option abbreviation in the options.

        Args:
            abbreviation ([type]): [description]

        Returns:
            [type]: [description]
        """
        if abbreviation in self._options.values():
            return True
        return False

    def solve(self, name: Union[AnyStr, None] = None) -> AnyStr:
        """Solve for abbreviation given a certain name. e.g.: center could return C

        Args:
            name (str, optional): Key to look for in the options for this Token.
                Defaults to None, which will return the default value set in the options
                for this Token.

        Raises:
            SolvingError: If Token is required and no value is passed.

            SolvingError: If given name is not found in options list.

        Returns:
            str: If Token is required, the same input value is returned

            str: If Token has options, the abbreviation for given name is returned

            str: If nothing is passed and Token has options, default option is returned.
        """
        if self.required and name:
            return name
        elif self.required and len(self._fallback):
            return self._fallback
        elif self.required and not name:
            raise TokenError(
                f"Token {self.name} is required. name parameter must be passed."
            )
        elif not self.required and name:
            if name not in self._options.keys():
                lower_match = [k.lower() for k in self._options.keys()]
                error_msg = (
                    f"Name '{name}' not found in Token '{self.name}'. "
                    f"Options: {', '.join(self._options.keys())}"
                )
                if name.lower() in lower_match:
                    error_msg = (
                        f"Check casing of '{name}' against token '{self.name}' "
                        f"options: {', '.join(self._options.keys())}"
                    )
                raise TokenError(error_msg)
            return self._options.get(name)
        elif not self.required and not name:
            return self._options.get(self.default)

    def parse(self, value: AnyStr) -> AnyStr:
        """Get metatada (origin) for given value in name. e.g.: L could return left

        Args:
            value (str): Name part to be parsed to the token origin

        Returns:
            str: Token origin for given value or value itself if token is required.
        """
        if self.required:
            return value
        elif not self.required and len(self._options) >= 1:
            for k, v in self._options.items():
                if v == value:
                    return k
        raise TokenError(
            f"Value '{value}' not found in Token '{self.name}'. "
            f"Options: {', '.join(self._options.values())}"
        )

    @property
    def required(self) -> bool:
        """
        Returns:
            [bool]: True if Token is required, False otherwise
        """
        return self.default is None

    @property
    def name(self) -> AnyStr:
        """
        Returns:
            [str]: Name of this Token
        """
        return self._name

    @name.setter
    def name(self, n: AnyStr):
        self._name = n

    @property
    def default(self) -> AnyStr:
        """If Token has options, one of them will be default. Either passed by the user,
        or simply the first found item in options.

        Returns:
            str: Default option value
        """
        if self._default is None and len(self._options) >= 1:
            self._default = sorted(list(self._options.keys()))[0]
        return self._default

    @default.setter
    def default(self, d: AnyStr):
        """
        Args:
            d (str): Value of the default option to be set
        """
        self._default = d

    @property
    def options(self) -> Dict:
        """
        Returns:
            [dict]: {"option_full_name":"abbreviation"}
        """
        return copy.deepcopy(self._options)

    @property
    def fallback(self) -> AnyStr:
        return self._fallback

    @fallback.setter
    def fallback(self, f: AnyStr):
        if self.required:
            self._fallback = f
        else:
            logger.warning(
                f"Token '{self.name}' has options, use {self.name}.default instead."
            )

    @property
    def nice_name(self) -> AnyStr:
        """
        Returns:
            [str]: Nice name of this Token
        """
        return self._nice_name

    @nice_name.setter
    def nice_name(self, n: AnyStr):
        self._nice_name = n


class TokenNumber(Serializable):
    def __init__(self, name: AnyStr, nice_name: AnyStr = ""):
        """Token for numbers with the ability to handle pure digits and version like strings
        (e.g.: v0025) with padding settings.

        In TokenNumber, options are limited to prefix, suffix and padding.

        Args:
            name (str): Name that best describes the TokenNumber, this will be used as a way
            to invoke the Token object.

            nice_name(str, optional): A more human readable name for the Token if needed for UI.
        """
        super(TokenNumber, self).__init__()
        self._name: AnyStr = name
        self._nice_name = name
        if len(nice_name):
            self._nice_name = nice_name
        self._default: int = 1
        self._options: Dict = {"prefix": "", "suffix": "", "padding": 3}

    def solve(self, number: int) -> AnyStr:
        """Solve for number with prefix, suffix and padding parameter found in the instance
            options. e.g.: 1 with padding 3, will return 001

        Args:
            number (int): Number to be solved for.

        Returns:
            str: The solved string to be used in the name
        """
        number_str = str(number).zfill(self.padding)
        return f"{self.prefix}{number_str}{self.suffix}"

    def parse(self, value: AnyStr) -> int:
        """Get metadata (number) for given value in name. e.g.: v0025 will return 25

        Args:
            value (str): String value taken from a name with digits in it.

        Returns:
            int: Number found in the given string.
        """
        if value.isdigit():
            return int(value)
        else:
            prefix_index = 0
            for each in value[::1]:
                if each.isdigit() and prefix_index == 0:
                    prefix_index = -1
                    break
                elif each.isdigit() and prefix_index > 0:
                    break
                prefix_index += 1

            suffix_index = 0
            for each in value[::-1]:
                if each.isdigit() and suffix_index == 0:
                    suffix_index = -1
                    break
                elif each.isdigit() and suffix_index > 0:
                    break
                suffix_index += 1

            if prefix_index != -1 and self.prefix != "":
                if value[prefix_index : len(self.prefix)] != self.prefix:  # noqa: E203
                    logger.warning(f"Prefix '{self.prefix}' not found in '{value}'")
            if suffix_index != -1 and self.suffix != "":
                if value[-suffix_index:] != self.suffix:
                    logger.warning(f"Suffix '{self.suffix}' not found in '{value}'")

            if prefix_index == -1 and suffix_index >= 0:
                return int(value[:-suffix_index])
            elif prefix_index >= 0 and suffix_index == -1:
                return int(value[prefix_index:])
            elif prefix_index >= 0 and suffix_index >= 0:
                return int(value[prefix_index:-suffix_index])

    @property
    def name(self) -> AnyStr:
        """
        Returns:
            [str]: Name of this Token
        """
        return self._name

    @name.setter
    def name(self, n: AnyStr):
        self._name = n

    @property
    def nice_name(self) -> AnyStr:
        """
        Returns:
            [str]: Nice name of this Token
        """
        return self._nice_name

    @nice_name.setter
    def nice_name(self, n: AnyStr):
        self._nice_name = n

    @property
    def default(self) -> AnyStr:
        return self._default

    @property
    def required(self) -> bool:
        return True

    @property
    def padding(self) -> int:
        return self._options.get("padding")

    @padding.setter
    def padding(self, p: int):
        if p <= 0:
            p = 1
        self._options["padding"] = int(p)

    @property
    def prefix(self) -> AnyStr:
        return self._options.get("prefix")

    @prefix.setter
    def prefix(self, this_prefix: AnyStr):
        if isinstance(this_prefix, str) and not this_prefix.isdigit():
            self._options["prefix"] = this_prefix
        else:
            logger.warning(f"Prefix must be a string: {this_prefix}")

    @property
    def suffix(self) -> AnyStr:
        return self._options.get("suffix")

    @suffix.setter
    def suffix(self, this_suffix: AnyStr):
        if isinstance(this_suffix, str) and not this_suffix.isdigit():
            self._options["suffix"] = this_suffix
        else:
            logger.warning(f"Suffix must be a string: {this_suffix}")

    @property
    def options(self) -> Dict:
        return copy.deepcopy(self._options)


def add_token(
    name: AnyStr, fallback: AnyStr = "", nice_name: AnyStr = "", **kwargs
) -> Token:
    """Add token to current naming session. If 'default' keyword argument is found,
    set it as default for the token instance.

    Args:
        name (str): Name that best describes the token, this will be used as a way
        to invoke the Token object.

        fallback (str, optional): Fallback value to use if token is required. Default is ""
        and will raise and error, making the token mandatory.

        nice_name(str, optional): A more human readable name for the Token if needed for UI.

        kwargs: Each argument following fallback is treated as an option for the
        new Token.

    Raises:
        TokenError: If explicitly passed default does not match with any of the options.

    Returns:
        Token: The Token object instance created for given name and fields.
    """
    token = Token(name)
    if len(nice_name):
        token.nice_name = nice_name
    for k, v in kwargs.items():
        if k == "default":
            continue
        token.add_option(k, v)
    if "default" in kwargs.keys():
        extract_default = copy.deepcopy(kwargs)
        del extract_default["default"]
        if kwargs.get("default") in extract_default.keys():
            token.default = kwargs.get("default")
        elif kwargs.get("default") in extract_default.values():
            for k, v in extract_default.items():
                if v == kwargs.get("default"):
                    token.default = k
                    break
        else:
            raise TokenError("Default value must match one of the options passed.")
    if len(fallback):
        if isinstance(fallback, str):
            token.fallback = fallback
        else:
            raise TokenError(f"Fallback must be a string. Got {type(fallback)}")

    _tokens[name] = token
    return token


def add_token_number(
    name,
    prefix: AnyStr = "",
    suffix: AnyStr = "",
    padding: int = 3,
    nice_name: AnyStr = "",
) -> TokenNumber:
    """Add token number to current naming session.

    Args:
        name (str): Name that best describes the token, this will be used as a way
        to invoke the TokenNumber object.

        nice_name(str, optional): A more human readable name for the TokenNumber if needed for UI.

        prefix (str, optional): Prefix for token number. Useful if you have to use 'v' as prefix
        for versioning for example.

        suffix (str, optional): Suffix for token number.

        padding (int, optional): Padding a numeric value with this leading number of zeroes.
        e.g.: 25 with padding 4 would be 0025

    Returns:
        TokenNumber: The TokenNumber object instance created for given name and fields.
    """
    token = TokenNumber(name)
    if len(nice_name):
        token.nice_name = nice_name
    token.prefix = prefix
    token.suffix = suffix
    token.padding = padding
    _tokens[name] = token
    return token


def remove_token(name: AnyStr) -> bool:
    """Remove Token or TokenNumber from current session.

    Args:
        name (str): The name of the token to be removed.

    Returns:
        bool: True if successful, False if a rule name was not found.
    """
    if has_token(name):
        del _tokens[name]
        return True
    return False


def has_token(name: AnyStr) -> bool:
    """Test if current session has a token with given name.

    Args:
        name (str): The name of the token to be looked for.

    Returns:
        bool: True if rule with given name exists in current session, False otherwise.
    """
    return name in _tokens.keys()


def update_token_name(old_name, new_name):
    """Update token name.

    Args:
        old_name (str): The current name of the token to update.

        new_name (str): The new name of the token to be updated.

    Returns:
        True if Token name was updated, False if another token
        has that name already or no current template with old_name was found.
    """
    if has_token(old_name) and not has_token(new_name):
        token_obj = _tokens.pop(old_name)
        token_obj.name = new_name
        _tokens[new_name] = token_obj
        if _tokens.get("_active") == old_name:
            _tokens["_active"] == new_name
        return True
    return False


def reset_tokens() -> bool:
    """Clears all rules created for current session.

    Returns:
        bool: True if clearing was successful.
    """
    _tokens.clear()
    return True


def get_token(name: AnyStr) -> Union[Token, TokenNumber, None]:
    """Gets Token or TokenNumber object with given name.

    Args:
        name (str): The name of the token to query.

    Returns:
        Rule: Token object instance for given name.
    """
    return _tokens.get(name)


def get_token_options(token_name):
    """Gets Token options for given token

    Args:
        ``token_name`` (str): The name of the token to query.

    Returns:
        [dict]: Token options. None if no token with given name was found,
        or token has no options.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        return token_obj.options
    return None


def get_token_default_option(token_name):
    """Gets Token default option for given token

    Args:
        ``token_name`` (str): The name of the token to query.

    Returns:
        [dict]: Token default option. None if no token with given name was found,
        or token has no options.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        option_dict = {token_obj.default: token_obj.options.get(token_obj.default)}
        return option_dict
    return None


def add_option_to_token(token_name, fullname, abbreviation):
    """Adds an option pair to this Token.

    Args:
        ``token_name`` (str): The name of the exisiting token.

        ``fullname`` (str): Full length name of the option.

        ``abbreviation`` (str): Abbreviation to be used when building the path.

    Returns:
        [bool]: True if successful. False otherwise.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        return token_obj.add_option(fullname, abbreviation)
    return False


def update_option_fullname_from_token(token_name, old_fullname, new_fullname):
    """Update an option fullname on this Token.

    Args:
        ``token_name`` (str): The name of the exisiting token.

        ``old_fullname`` (str): Old full length name of the option.

        ``new_fullname`` (str): New full length name of the option.

        ``abbreviation`` (str): Abbreviation to be used when building the path.

    Returns:
        [bool]: True if successful. False otherwise.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        if token_obj.has_option_fullname(old_fullname):
            abbreviation = token_obj.options.get(old_fullname)
            token_obj.remove_option(old_fullname)
            return token_obj.add_option(new_fullname, abbreviation)
    return False


def update_option_abbreviation_from_token(token_name, fullname, abbreviation):
    """Update an option abbreviation on this Token.

    Args:
        ``token_name`` (str): The name of the exisiting token.

        ``fullname`` (str): Full length name of the option.

        ``abbreviation`` (str): Abbreviation to be used when building the path.

    Returns:
        [bool]: True if successful. False otherwise.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        if token_obj.has_option_fullname(fullname):
            return token_obj.update_option(fullname, abbreviation)
    return False


def remove_option_from_token(token_name, fullname):
    """Remove an option on given token.

    Args:
        ``token_name`` (str): The name of the exisiting token.

        ``fullname`` (str): Full length name of the option

    Returns:
        [bool]: True if successful. False otherwise.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        if token_obj.has_option_fullname(fullname):
            return token_obj.remove_option(fullname)
    return False


def has_option_fullname(token_name, fullname):
    """Looks for given option full name in the given token options.

    Args:
        ``fullname`` (str): Full name of the option

    Returns:
        [bool]: True if found. False otherwise.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        return token_obj.has_option_fullname(fullname)
    return False


def has_option_abbreviation(token_name, abbreviation):
    """Looks for given option abbreviation in the given token options.

    Args:
        ``abbreviation`` (str): Abbreviation

    Returns:
        [bool]: True if found. False otherwise.
    """
    if has_token(token_name):
        token_obj = get_token(token_name)
        return token_obj.has_option_abbreviation(abbreviation)
    return False


def get_tokens() -> Dict:
    """Get all Token and TokenNumber objects for current session.

    Returns:
        dict: {token_name:token_object}
    """
    return _tokens


def validate_tokens():
    not_valid = list()
    for name, token_obj in get_tokens().items():
        if not token_obj.required and len(token_obj.options) == 0:
            not_valid.append(name)
    if len(not_valid) >= 1:
        raise TokenError(
            f"Tokens {', '.join(not_valid)}: Not required tokens must "
            "have at least one option (fullname=abbreviation)."
        )


def save_token(name: AnyStr, directory: Path) -> bool:
    """Saves given token serialized to specified location.

    Args:
        name (str): The name of the token to be saved.
        filepath (str): Path location to save the token.

    Returns:
        bool: True if successful, False if rule wasn't found in current session.
    """
    token = get_token(name)
    if not token:
        return False
    file_name = f"{name}.token"
    filepath = directory / file_name
    with open(filepath, "w") as fp:
        json.dump(token.data(), fp)
    return True


def load_token(filepath: Path) -> bool:
    """Load token from given location and create Token or TokenNumber object in
    memory to work with it.

    Args:
        filepath (str): Path to existing .token file location

    Returns:
        bool: True if successful, False if .token wasn't found.
    """
    if not filepath.is_file():
        return False
    try:
        with open(filepath) as fp:
            data = json.load(fp)
    except Exception:
        return False
    class_name = data.get("_Serializable_classname")
    logger.debug(f"Loading token type: {class_name}")
    token_class = globals().get(class_name)
    if token_class is None:
        return False
    token = getattr(token_class, "from_data")(data)
    if token:
        _tokens[token.name] = token
        return True
    return False



import re
import json
import traceback
import functools
from copy import deepcopy
from pathlib import Path
from collections import defaultdict
from typing import Dict, AnyStr, Union, Tuple

from vfxnaming.serialize import Serializable
from vfxnaming.tokens import get_token, TokenNumber
from vfxnaming.logger import logger
from vfxnaming.error import ParsingError, SolvingError, RuleError

_rules = {"_active": None}


class Rule(Serializable):
    """Each rule is managed by an instance of this class. Fields exist for each
    Token and Separator used in the rule definition.

    Args:
        ``name`` (str): Name that best describes the rule, this will be used as a way
        to query the Rule object.

        ``pattern`` (str): The template pattern to use, which uses existing Tokens.
        e.g.: '{side}_{region}_{side}_{region}.png'

        ``nice_name`` (str, optional): A more human readable name for the Token if needed for UI.

        ``anchor``: ([Rule.ANCHOR_START, Rule.ANCHOR_END, Rule.ANCHOR_BOTH], optional):
        For parsing, regex matching will look for a match from this Anchor. If a
        pattern is anchored to the start, it requires the start of a passed path to
        match the pattern. Defaults to ANCHOR_START.
    """

    __FIELDS_REGEX = re.compile(r"{(.+?)}")
    __PATTERN_SEPARATORS_REGEX = re.compile(
        r"(}[_\-\.:\|/\\]{|[_\-\.:\|/\\]{|}[_\-\.:\|/\\])"
    )
    __SEPARATORS_REGEX = re.compile(r"[_\-\.:\|/\\]")
    __RULE_REFERENCE_REGEX = re.compile(r"{@(?P<reference>.+?)}")
    __AT_CODE = "_FXW_"
    ANCHOR_START, ANCHOR_END, ANCHOR_BOTH = (1, 2, 3)

    def __init__(self, name, pattern, anchor=ANCHOR_START, nice_name: AnyStr = ""):
        super(Rule, self).__init__()
        self._name: str = name
        self._nice_name: str = name
        if len(nice_name):
            self._nice_name = nice_name
        self._pattern: str = pattern
        self._anchor: int = anchor
        self._regex: re.Pattern = self.__build_regex()

    def data(self) -> Dict:
        """Collect all data for this object instance.

        Returns:
            dict: {attribute:value}
        """
        retval = dict()
        retval["_name"] = self._name
        retval["_nice_name"] = self._nice_name
        retval["_pattern"] = self._pattern
        retval["_anchor"] = self._anchor
        retval["_Serializable_classname"] = type(self).__name__
        retval["_Serializable_version"] = "1.0"
        return retval

    @classmethod
    def from_data(cls, data) -> "Rule":
        """Create object instance from give data. Used by Rule,
        Token, Separator to create object instances from disk saved data.

        Args:
            data (dict): {attribute:value}

        Returns:
            Serializable: Object instance for Rule, Token or Separator.
        """
        # Validation
        if data.get("_Serializable_classname") != cls.__name__:
            return None
        del data["_Serializable_classname"]
        if data.get("_Serializable_version") is not None:
            del data["_Serializable_version"]

        if not data.get("_name"):
            raise RuleError(f"Rule name is required but was not found in {data}")

        if not data.get("_pattern"):
            raise RuleError(f"Rule pattern is required but was not found in {data}")

        this = cls(
            data.get("_name"),
            data.get("_pattern"),
            data.get("_anchor", cls.ANCHOR_START),
            data.get("_nice_name", ""),
        )
        return this

    def solve(self, **values) -> AnyStr:
        """Given arguments are used to build a name.

        Raises:
            SolvingError: Arguments passed do not match with rule fields.

        Returns:
            str: A string with the resulting name.
        """
        result = None

        try:
            result = self.__digits_pattern().format(**values)
        except KeyError as why:
            raise SolvingError(
                f"Arguments passed do not match with naming rule fields {self._pattern}\n{why}"
            )

        return result

    def parse(self, name: AnyStr) -> Union[Dict, None]:
        """Build and return dictionary with keys as tokens and values as given names.

        If your rule uses the same token more than once, the returned dictionary keys
        will have the token name and an incremental digit next to them so they can be
        differentiated.

        Args:
            name (str): Name string e.g.: C_helmet_001_MSH

        Returns:
            dict: A dictionary with keys as tokens and values as given name parts.
            e.g.: {'side':'C', 'part':'helmet', 'number': 1, 'type':'MSH'}
        """
        expected_separators = self.__PATTERN_SEPARATORS_REGEX.findall(self._pattern)
        if len(expected_separators) <= 0:
            logger.warning(
                f"No separators used for rule '{self.name}', parsing is not possible."
            )
            return None
        name_separators = self.__SEPARATORS_REGEX.findall(name)
        if len(expected_separators) <= len(name_separators):
            parsed = {}
            regex = self.__build_regex()
            match = regex.search(name)
            if match:
                name_parts = sorted(match.groupdict().items())
                name_parts_str = ", ".join(
                    [f"('{k[:-3]}': '{v}')" for k, v in name_parts]
                )
                logger.debug(f"Name parts: {name_parts_str}")
                repeated_fields = {}
                for each in self.fields:
                    if each not in repeated_fields.keys():
                        if self.fields.count(each) > 1:
                            repeated_fields[each] = 1
                if repeated_fields:
                    logger.debug(f"Repeated tokens: {', '.join(repeated_fields.keys())}")

                for key, value in name_parts:
                    # Strip number that was added to make group name unique
                    token_name = key[:-3]
                    token = get_token(token_name)
                    if not token:
                        continue
                    if token_name in repeated_fields.keys():
                        counter = repeated_fields.get(token_name)
                        repeated_fields[token_name] = counter + 1
                        token_name = f"{token_name}{counter}"
                    parsed[token_name] = token.parse(value)
            return parsed
        else:
            raise ParsingError(
                f"Separators count mismatch between given name '{name}':'{len(name_separators)}' "
                f"and rule's pattern '{self._pattern}':'{len(expected_separators)}'."
            )

    def validate(self, name: AnyStr, strict: bool = False, **validate_values) -> bool:  # noqa: C901
        """Validate if given name matches the rule pattern.

        Args:
            name (str): Name string e.g.: C_helmet_001_MSH

        Returns:
            bool: True if name matches the rule pattern, False otherwise.
        """
        expected_separators = self.__PATTERN_SEPARATORS_REGEX.findall(self._pattern)
        if len(expected_separators) <= 0:
            logger.warning(
                f"No separators used for rule '{self.name}', parsing is not possible."
            )
            return False

        name_separators = self.__SEPARATORS_REGEX.findall(name)
        if not strict and len(expected_separators) > len(name_separators):
            logger.warning(
                f"Separators count mismatch between given name '{name}':'{len(name_separators)}' "
                f"and rule's pattern '{self._pattern}':'{len(expected_separators)}'."
            )
            return False
        elif strict and len(expected_separators) != len(name_separators):
            logger.warning(
                f"Separators count mismatch between given name '{name}':'{len(name_separators)}' "
                f"and rule's pattern '{self._pattern}':'{len(expected_separators)}'."
            )
            return False

        regex = self.__build_regex(strict)
        match = regex.search(name)
        if not match:
            logger.warning(f"Name {name} does not match rule pattern '{self._pattern}'")
            return False

        match_dict = match.groupdict()
        name_parts = sorted(match_dict.items())
        repeated_fields = {}
        for each in self.fields:
            if each not in repeated_fields.keys():
                if self.fields.count(each) > 1:
                    repeated_fields[each] = 1
        if repeated_fields:
            logger.debug(f"Repeated tokens: {', '.join(repeated_fields.keys())}")

        # Validate values passed by the user
        if len(validate_values):
            for key, value in name_parts:
                # Strip number that was added to make group name unique
                token_name = key[:-3]
                token = get_token(token_name)
                if not token:
                    continue
                if token_name in repeated_fields.keys():
                    counter = repeated_fields.get(token_name)
                    repeated_fields[token_name] = counter + 1
                    token_name = f"{token_name}{counter}"

                if token_name not in validate_values.keys():
                    continue

                given_value = validate_values.get(token_name)
                if value != given_value:
                    logger.warning(
                        f"Token '{token_name}' value '{value}' does not match '{given_value}'"
                    )
                    if value.lower() == given_value.lower():
                        logger.warning(
                            f"Token '{token_name}' value '{value}' has "
                            f"casing mismatches with '{given_value}'"
                        )
                    return False

        name_parts_str = ", ".join([f"('{k[:-3]}': '{v}')" for k, v in name_parts])
        logger.debug(f"Name parts: {name_parts_str}")

        has_tokens_with_options = False
        for key, value in name_parts:
            # Strip number that was added to make group name unique
            token_name = key[:-3]
            token = get_token(token_name)
            if not token.required:
                has_tokens_with_options = True
                break
        # If we don't have tokens with options this match is already valid
        if not has_tokens_with_options:
            return True

        matching_options = True
        for key, value in name_parts:
            # Strip number that was added to make group name unique
            token_name = key[:-3]
            token = get_token(token_name)
            if not token.required:
                if not (
                    token.has_option_fullname(value)
                    or token.has_option_abbreviation(value)
                ):
                    logger.warning(f"Token {token_name} has no option {value}")
                    matching_options = False
            if isinstance(token, TokenNumber):
                if len(token.suffix):
                    if not value.endswith(token.suffix):
                        logger.warning(
                            f"Token {token_name}: {value} must end with {token.suffix}"
                        )
                        matching_options = False
                if len(token.prefix):
                    if not value.startswith(token.prefix):
                        logger.warning(
                            f"Token {token_name}: {value} must end with {token.prefix}"
                        )
                        matching_options = False
                digits = value[len(token.prefix) : len(token.suffix) * -1]  # noqa: E203
                if not len(token.suffix):
                    digits = value[len(token.prefix) :]  # noqa: E203
                if not digits.isdigit():
                    logger.warning(
                        f"Token {token_name}: {value} must be digits with "
                        f"prefix '{token.prefix}' and suffix '{token.suffix}'"
                    )
                    matching_options = False
                    continue
                hash_str = "#" * token.padding
                limit = int(hash_str.replace("#", "9"))
                if len(digits) != token.padding and int(digits) <= limit:
                    logger.warning(
                        f"Token {token_name}: {value} must have {token.padding} digits"
                    )
                    if int(digits) > limit:
                        logger.debug(
                            f"Token {token_name}: {value} is higher than {limit}. "
                            "Consider increasing padding."
                        )
                    matching_options = False

        return matching_options

    def __build_regex(self, strict: bool = False) -> re.Pattern:
        # ? Taken from Lucidity by Martin Pengelly-Phillips
        # Escape non-placeholder components
        expression = re.sub(
            r"(?P<placeholder>{(.+?)(:(\\}|.)+?)?})|(?P<other>.+?)",
            self.__escape,
            self.expanded_pattern(),
        )
        # Replace placeholders with regex pattern
        expression = re.sub(
            r"{(?P<placeholder>.+?)(:(?P<expression>(\\}|.)+?))?}",
            functools.partial(self.__convert, placeholder_count=defaultdict(int)),
            expression,
        )

        if self._anchor is not None:
            if bool(self._anchor & self.ANCHOR_START):
                expression = f"^{expression}"

            if bool(self._anchor & self.ANCHOR_END):
                expression = f"{expression}$"
        # Compile expression
        try:
            if strict:
                compiled = re.compile(expression)
            else:
                compiled = re.compile(expression, re.IGNORECASE)
        except re.error as error:
            if any(
                [
                    "bad group name" in str(error),
                    "bad character in group name" in str(error),
                ]
            ):
                raise ValueError("Placeholder name contains invalid characters.")
            else:
                message = f"Invalid pattern.\n{traceback.format_exc()}"
                raise ValueError(message)

        return compiled

    def __convert(self, match: re.Match, placeholder_count: int) -> AnyStr:
        """Return a regular expression to represent *match*.

        ``placeholder_count`` should be a ``defaultdict(int)`` that will be used to
        store counts of unique placeholder names.

        """
        # ? Taken from Lucidity by Martin Pengelly-Phillips
        placeholder_name = match.group("placeholder")

        # Support at symbol (@) as referenced template indicator. Currently,
        # this symbol not a valid character for a group name in the standard
        # Python regex library. Rather than rewrite or monkey patch the library
        # work around the restriction with a unique identifier.
        placeholder_name = placeholder_name.replace("@", self.__AT_CODE)

        # The re module does not support duplicate group names. To support
        # duplicate placeholder names in templates add a unique count to the
        # regular expression group name and strip it later during parse.
        placeholder_count[placeholder_name] += 1
        placeholder_name += "{0:03d}".format(placeholder_count[placeholder_name])

        expression = match.group("expression")
        if expression is None:
            expression = r"[\w_.\-/:]+"

        # Un-escape potentially escaped characters in expression.
        expression = expression.replace("{", "{").replace("}", "}")

        return r"(?P<{0}>{1})".format(placeholder_name, expression)

    def expanded_pattern(self):
        """Return pattern with all referenced rules expanded recursively.

        Returns:
            [str]: Pattern with all referenced rules expanded recursively.
        """
        # ? Taken from Lucidity by Martin Pengelly-Phillips
        return self.__RULE_REFERENCE_REGEX.sub(self.__expand_reference, self.pattern)

    def expanded_pattern_validation(self, pattern):
        """Return pattern with all referenced rules expanded recursively from a given pattern

        Args:
            ``pattern`` (str): Pattern string.
            e.g.: "{@rule_base}/{token_1}/{token_2}/inmutable_folder/"

        Returns:
            [str]: Pattern with all referenced rules expanded recursively.
        """
        # ? Taken from Lucidity by Martin Pengelly-Phillips
        return self.__RULE_REFERENCE_REGEX.sub(self.__expand_reference, pattern)

    def __expand_reference(self, match):
        """Expand reference represented by *match*.

        Args:
            match (str): Template name to look for in repo.

        Raises:
            TemplateError: If pattern contains a reference that cannot be
            resolved.

        Returns:
            [str]: Expanded reference pattern
        """
        # ? Taken from Lucidity by Martin Pengelly-Phillips
        reference = match.group("reference")

        rule = get_rule(reference)
        if rule is None:
            raise RuleError(
                "Failed to find reference {} in current repo.".format(reference)
            )

        return rule.expanded_pattern()

    def __escape(self, match: re.Match) -> AnyStr:
        """Escape matched 'other' group value."""
        # ? Taken from Lucidity by Martin Pengelly-Phillips
        groups = match.groupdict()
        if groups.get("other"):
            return re.escape(groups.get("other"))
        return groups.get("placeholder")

    def __digits_pattern(self):
        # * This accounts for those cases where a token is used more than once in a rule
        digits_pattern = deepcopy(self.expanded_pattern())
        for each in list(set(self.fields)):
            # ? This is going to be a bit more difficult to handle when nesting templates
            # ? due to the . character not being contemplated by the pattern
            regex_pattern = re.compile("{{{0}}}".format(each))
            indexes = [match.end() for match in regex_pattern.finditer(digits_pattern)]
            repetitions = len(indexes)
            if repetitions > 1:
                i = 0
                for match in sorted(indexes, reverse=True):
                    digits_pattern = (
                        f"{digits_pattern[: match - 1]}"
                        f"{str(repetitions - i)}"
                        f"{digits_pattern[match - 1 :]}"
                    )
                    i += 1
        logger.debug(f"Digits pattern to account for repeated fields: {digits_pattern}")
        return digits_pattern

    @property
    def pattern(self) -> AnyStr:
        """
        Returns:
            [str]: Pattern that this Rule was initialized with.
        """
        return self._pattern

    @pattern.setter
    def pattern(self, pattern):
        """
        Some times we need to change the pattern dinamically, at runtime.
        """
        if pattern == "":
            logger.error(f"Pattern cannot be empty for rule: {self.name}")
            return
        self._pattern = pattern

    @property
    def fields(self) -> Tuple:
        """
        Returns:
            [tuple]: Tuple of all Tokens found in this Rule's pattern
        """
        return tuple(self.__FIELDS_REGEX.findall(self.expanded_pattern()))

    @property
    def anchor(self):
        """
        Returns:
            [int]: Rule.ANCHOR_START, Rule.ANCHOR_END, Rule.ANCHOR_BOTH = (1, 2, 3)
        """
        return self._anchor

    @property
    def name(self) -> AnyStr:
        return self._name

    @name.setter
    def name(self, n: str):
        if n == "":
            logger.error(f"Name cannot be empty for rule: {self._pattern}")
        self._name = n

    @property
    def nice_name(self) -> AnyStr:
        """
        Returns:
            [str]: Nice name of this Rule
        """
        return self._nice_name

    @nice_name.setter
    def nice_name(self, n: AnyStr):
        self._nice_name = n


def add_rule(
    name: str, pattern: str, anchor=Rule.ANCHOR_START, nice_name: str = ""
) -> Union[Rule, None]:
    """Add rule to current naming session. If no active rule is found, it adds
    the created one as active by default.

    Args:
        ``name`` (str): Name that best describes the rule, this will be used as a way
        to invoke the Rule object.

        ``nice_name`` (str, optional): A more human readable name for the Rule if needed for UI.

        ``pattern`` (str): The template pattern to use, which uses existing Tokens.
        e.g.: '{side}_{region}_{side}_{region}.png'

        ``anchor``: ([Rule.ANCHOR_START, Rule.ANCHOR_END, Rule.ANCHOR_BOTH], optional):
        For parsing, regex matching will look for a match from this Anchor. If a
        pattern is anchored to the start, it requires the start of a passed path to
        match the pattern. Defaults to ANCHOR_START.

    Returns:
        Rule: The Rule object instance created for given name and fields. None if
    """
    if pattern == "":
        logger.error(f"Pattern cannot be empty for rule: {name}")
        return
    if name == "":
        logger.error(f"Name cannot be empty for rule: {pattern}")
        return
    if anchor not in [Rule.ANCHOR_START, Rule.ANCHOR_END, Rule.ANCHOR_BOTH]:
        logger.error(f"Invalid anchor value for rule: {name}")
        return
    rule = Rule(name, pattern, anchor)
    if len(nice_name):
        rule.nice_name = nice_name
    _rules[name] = rule
    if get_active_rule() is None:
        set_active_rule(name)
        logger.debug(f"No active rule found, setting this one as active: {name}")
    return rule


def remove_rule(name: AnyStr) -> bool:
    """Remove Rule from current session.

    Args:
        name (str): The name of the rule to be removed.

    Returns:
        bool: True if successful, False if a rule name was not found.
    """
    if has_rule(name):
        del _rules[name]
        return True
    return False


def has_rule(name: AnyStr) -> bool:
    """Test if current session has a rule with given name.

    Args:
        name (str): The name of the rule to be looked for.

    Returns:
        bool: True if rule with given name exists in current session, False otherwise.
    """
    return name in _rules.keys()


def update_rule_name(old_name, new_name):
    """Update rule name.

    Args:
        old_name (str): The current name of the rule to update.

        new_name (str): The new name of the rule to be updated.

    Returns:
        True if Rule name was updated, False if another rule
        has that name already or no current rule with old_name was found.
    """
    if has_rule(old_name) and not has_rule(new_name):
        rule_obj = _rules.pop(old_name)
        rule_obj.name = new_name
        _rules[new_name] = rule_obj
        if _rules.get("_active") == old_name:
            _rules["_active"] == new_name
        return True
    return False


def reset_rules() -> bool:
    """Clears all rules created for current session.

    Returns:
        bool: True if clearing was successful.
    """
    _rules.clear()
    _rules["_active"] = None
    return True


def get_active_rule() -> Rule:
    """Get currently active rule for the session. This is the rule
    that will be used to parse and solve from.

    Returns:
        Rule: Rule object instance for currently active Rule.
    """
    name = _rules.get("_active")
    return _rules.get(name)


def set_active_rule(name: AnyStr) -> bool:
    """Sets given rule as active for the session. This it the rule that's
    being used to parse and solve from.

    Args:
        name (str): The name of the rule to be activated.

    Returns:
        bool: True if successful, False otherwise.
    """
    if has_rule(name):
        _rules["_active"] = name
        return True
    return False


def validate_rule_pattern(name):
    """Validates rule pattern is not empty.

    Args:
        name (str): The name of the rule to validate its pattern.

    Returns:
        bool: True if successful, False otherwise.
    """
    if has_rule(name):
        rule_obj = get_rule(name)
        if len(rule_obj.pattern) >= 1:
            return True
    return False


def validate_rules():
    not_valid = []
    for name, rule in get_rules().items():
        if not validate_rule_pattern(name):
            not_valid.append(name)
    if len(not_valid) >= 1:
        raise RuleError(f"Rules {', '.join(not_valid)}: Patterns are not valid.")


def get_rule(name: AnyStr) -> Rule:
    """Gets Rule object with given name.

    Args:
        name (str): The name of the rule to query.

    Returns:
        Rule: Rule object instance for given name.
    """
    return _rules.get(name)


def get_rules() -> Dict:
    """Get all Rule objects for current session.

    Returns:
        dict: {rule_name:Rule}
    """
    rules_copy = deepcopy(_rules)
    del rules_copy["_active"]
    return rules_copy


def save_rule(name: AnyStr, directory: Path) -> bool:
    """Saves given rule serialized to specified location.

    Args:
        ``name`` (str): The name of the rule to be saved.

        ``directory`` (str): Path location to save the rule.

    Returns:
        bool: True if successful, False if rule wasn't found in current session.
    """
    rule = get_rule(name)
    if not rule:
        return False
    file_name = f"{name}.rule"
    filepath = directory / file_name
    with open(filepath, "w") as fp:
        json.dump(rule.data(), fp)
    return True


def load_rule(filepath: Path) -> bool:
    """Load rule from given location and create Rule object in memory to work with it.

    Args:
        filepath (str): Path to existing .rule file location

    Returns:
        bool: True if successful, False if .rule wasn't found.
    """
    if not filepath.is_file():
        return False
    try:
        with open(filepath) as fp:
            data = json.load(fp)
    except Exception:
        return False
    new_rule = Rule.from_data(data)
    if new_rule:
        _rules[new_rule.name] = new_rule
        return True
    return False



# coding=utf-8
# MIT License
# Copyright (c) 2017 Cesar Saez and modified by Chris Granados- Xian
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import re
import json
import traceback
import shutil
import vfxnaming.rules as rules
import vfxnaming.tokens as tokens
from pathlib import Path
from typing import AnyStr, Dict, Union, Iterable

from vfxnaming.logger import logger
from vfxnaming.error import SolvingError, RepoError


NAMING_REPO_ENV = "NAMING_REPO"


def parse(name: AnyStr) -> Dict:
    """Get metadata from a name string recognized by the currently active rule.

    -For rules with repeated tokens:

    If your rule uses the same token more than once, the returned dictionary keys
    will have the token name and an incremental digit next to them so they can be
    differentiated.

    Args:
        name (str): Name string e.g.: C_helmet_001_MSH

    Returns:
        dict: A dictionary with keys as tokens and values as given name parts.
        e.g.: {'side':'C', 'part':'helmet', 'number': 1, 'type':'MSH'}
    """
    rule = rules.get_active_rule()
    return rule.parse(name)


def solve(*args, **kwargs) -> AnyStr:
    """Given arguments are used to build a name following currently active rule.

    -For rules with repeated tokens:

    If your rule uses the same token more than once, pass arguments with the token
    name and add an incremental digit

    i.e.: side1='C', side2='R'

    If your rule uses the same token more than once, you can also pass a single
    instance of the argument and it'll be applied to all repetitions.

    i.e.: side='C'

    If your rule uses the same token more than once, you can ignore one of the repetitions,
    and the solver will use the default value for that token.

    i.e.: side1='C', side4='L'

    Raises:
        SolvingError: A required token was passed as None to keyword arguments.
        SolvingError: Missing argument for one field in currently active rule.

    Returns:
        str: A string with the resulting name.
    """
    rule: rules.Rule = rules.get_active_rule()
    # * This accounts for those cases where a token is used more than once in a rule
    repeated_fields = dict()
    for each in rule.fields:
        if each not in repeated_fields.keys():
            if rule.fields.count(each) > 1:
                repeated_fields[each] = 1
    fields_with_digits = list()
    for each in rule.fields:
        if each in repeated_fields.keys():
            counter = repeated_fields.get(each)
            repeated_fields[each] = counter + 1
            fields_with_digits.append(f"{each}{counter}")
        else:
            fields_with_digits.append(each)
    values = {}
    i = 0
    fields_inc = 0
    for f in fields_with_digits:
        token = tokens.get_token(rule.fields[fields_inc])
        if token:
            # Explicitly passed as keyword argument
            if kwargs.get(f) is not None:
                values[f] = token.solve(kwargs.get(f))
                fields_inc += 1
                continue
            # Explicitly passed as keyword argument without repetitive digits
            # Use passed argument for all field repetitions
            elif kwargs.get(rule.fields[fields_inc]) is not None:
                values[f] = token.solve(kwargs.get(rule.fields[fields_inc]))
                fields_inc += 1
                continue
            elif token.required and kwargs.get(f) is None and len(args) == 0:
                if isinstance(token, tokens.Token):
                    if len(token.fallback):
                        values[f] = token.fallback
                        fields_inc += 1
                        continue
                else:
                    raise SolvingError(
                        f"Token {token.name} is required but was not passed."
                    )
            # Not required and not passed as keyword argument
            elif not token.required and kwargs.get(f) is None:
                values[f] = token.solve()
                fields_inc += 1
                continue
            # Implicitly passed as positional argument
            try:
                values[f] = token.solve(args[i])
                i += 1
                fields_inc += 1
                continue
            except IndexError as why:
                raise SolvingError(f"Missing argument for field '{f}'\n{why}")
    logger.debug(f"Solving rule '{rule.name}' with values {values}")
    return rule.solve(**values)


def validate(  # noqa: C901
    name: AnyStr, with_rules: Iterable[str] = [], strict: bool = False, **kwargs
) -> Iterable[rules.Rule]:
    """Validates a name string against the currently active rule if no rules are passed or
    against the list of specific rules passed in with_rules.
    It also validates its tokens if passed as keyword arguments.

    -For rules with repeated tokens:

    If your rule uses the same token more than once, pass arguments with the token
    name and add an incremental digit

    i.e.: side1='C', side2='R'

    If your rule uses the same token more than once, you can also pass a single
    instance of the argument and it'll be applied to all repetitions.

    i.e.: side='C'

    If your rule uses the same token more than once, you can ignore one of the repetitions,
    and the solver will use the default value for that token.

    i.e.: side1='C', side4='L'

    Args:
        name (str): Name string e.g.: C_helmet_001_MSH

        with_rules (list, optional): List of rule names to validate against. Defaults to [].

        strict (bool, optional): If False, it'll try to accept casing mismatches.

        kwargs (dict): Keyword arguments with token names and values.

    Returns:
        list: List of validated rules. Empty list if no rule could be validated.
    """
    previously_active_rule = rules.get_active_rule()
    if not len(with_rules):
        with_rules = [previously_active_rule.name]
    validated: Iterable[rules.Rule] = []
    for with_rule in with_rules:
        rule = rules.get_rule(with_rule)
        if not rule:
            logger.warning(f"Rule {with_rule} not found.")
            continue

        rules.set_active_rule(with_rule)
        # * This accounts for those cases where a token is used more than once in a rule
        repeated_fields = dict()
        for each in rule.fields:
            if each not in repeated_fields.keys():
                if rule.fields.count(each) > 1:
                    repeated_fields[each] = 1
        fields_with_digits = list()
        for each in rule.fields:
            if each in repeated_fields.keys():
                counter = repeated_fields.get(each)
                repeated_fields[each] = counter + 1
                fields_with_digits.append(f"{each}{counter}")
            else:
                fields_with_digits.append(each)
        values = {}
        fields_inc = 0
        for f in fields_with_digits:
            token = tokens.get_token(rule.fields[fields_inc])
            if token:
                # Explicitly passed as keyword argument
                if kwargs.get(f) is not None:
                    values[f] = token.solve(kwargs.get(f))
                    fields_inc += 1
                    continue
                # Explicitly passed as keyword argument without repetitive digits
                # Use passed argument for all field repetitions
                elif kwargs.get(rule.fields[fields_inc]) is not None:
                    values[f] = token.solve(kwargs.get(rule.fields[fields_inc]))
                    fields_inc += 1
                    continue
                fields_inc += 1
        logger.debug(f"Validating rule '{rule.name}' with values {values}")
        validation = rule.validate(name, strict, **values)
        if validation:
            rules.set_active_rule(previously_active_rule)
            validated.append(rule)
    rules.set_active_rule(previously_active_rule)
    if not len(validated):
        logger.warning(
            f"Could not validate {name} with any of the given "
            f"rules {', '.join([rule for rule in with_rules])}."
        )
    else:
        logger.info(
            f"Name {name} validated with rules: {', '.join([rule.name for rule in validated])}."
        )
    return validated


def validate_repo(repo: Path) -> bool:
    """Valides repo by checking if it contains a vfxnaming.conf file.

    Args:
        repo (Path): Repo dir

    Returns:
        bool: True if valid, False otherwise.
    """
    config_file = repo / "vfxnaming.conf"
    if not config_file.exists():
        return False
    return True


def validate_tokens_and_referenced_rules(pattern: str) -> bool:
    """Validate if the pattern uses tokens and rules that are defined in the current session.

    Args:
        pattern (str): Naming pattern to validate.

    Returns:
        bool: True if successful, False otherwise.
    """
    valid = True

    regex = re.compile(r"{(?P<placeholder>.+?)(:(?P<expression>(\\}|.)+?))?}")
    matches = regex.finditer(pattern)

    all_rules = list(rules.get_rules().keys())
    all_tokens = list(tokens.get_tokens().keys())

    rules_used = []
    tokens_used = []
    for match in matches:
        match_text = match.group(1)
        if match_text.startswith("@"):
            rules_used.append(match_text.replace("@", ""))
        else:
            tokens_used.append(match_text)

    for rule in rules_used:
        if rule not in all_rules:
            valid = False
            break

    for token in tokens_used:
        if token not in all_tokens:
            valid = False
            break

    return valid


def get_repo(force_repo: Union[Path, str] = None) -> Path:
    """Get the path to a folder structures repo.

    Path is looked for in these places and with the given priority:

        1- ``force_repo`` parameter

        2- Environment variable: NAMING_REPO

        3- User config file: C:/Users/xxxxx/.CGXTools/vfxnaming/config.json

        4- Dev config file: __file__/cfg/config.json

    In both config.json files the key it looks for is 'vfxnaming_repo'

    If a root path is passed as ``force_repo`` parameter, then it'll
    return the same path but first checks it actually exists.

    Keyword Arguments:
        ``force_repo`` {Path} -- Use this path instead of looking for
        pre-configured ones (default: {None})

    Raises:
        RepoError: Given repo directory does not exist.

        RepoError: Config file for vfxnmaing library couldn't be found.

    Returns:
        [Path] -- Root path
    """
    # Env level
    env_root = Path(os.environ.get(NAMING_REPO_ENV))

    # User level
    user_cfg_path = Path.expanduser("~") / ".CGXTools/vfxnaming/config.json"
    user_config = {}
    if user_cfg_path.exists():
        with open(user_cfg_path) as fp:
            user_config = json.load(fp)
    user_root = user_config.get("vfxnaming_repo")

    root = env_root or user_root
    if force_repo:
        root = force_repo

    if not validate_repo(root):
        raise RepoError(
            f"VFXNaming repo {root} is not valid, missing vfxnaming.conf file."
        )

    if root.exists():
        logger.debug(f"VFXNaming repo: {root}")
        return root

    raise RepoError(f"VFXNaming repo directory doesn't exist: {root}")


def save_session(repo: Union[Path, None] = None, override=True):
    """Save rules, tokens and config files to the repository.

    Raises:
        RepoError: Repository directory could not be created or is not valid.

        TokenError: Not required tokens must have at least one option (fullname=abbreviation).

        TemplateError: Template patterns are not valid.

    Args:
        ``repo`` (str, optional): Path to a repository. Defaults to None.

        ``override`` (bool, optional): If True, it'll remove given directory and recreate it.

    Returns:
        [bool]: True if saving session operation was successful.
    """
    # Validations
    rules.validate_rules()
    tokens.validate_tokens()

    repo = repo or get_repo()
    if override:
        try:
            shutil.rmtree(repo)
        except (IOError, OSError) as why:
            raise RepoError(why, traceback.format_exc())
    if not repo.exists():
        try:
            os.mkdir(repo)
        except (IOError, OSError) as why:
            raise RepoError(why, traceback.format_exc())

    # Save tokens
    for name, token in tokens.get_tokens().items():
        logger.debug(f"Saving token: {name} in {repo}")
        tokens.save_token(name, repo)
    # Save rules
    for name, template in rules.get_rules().items():
        if not isinstance(template, rules.Rule):
            continue
        logger.debug(f"Saving template: {name} in {repo}")
        rules.save_rule(name, repo)
    # extra configuration
    active = rules.get_active_rule()
    config = {"set_active_rule": active.name if active else None}
    filepath = os.path.join(repo, "vfxnaming.conf")
    logger.debug(f"Saving active rule: {active.name} in {filepath}")
    with open(filepath, "w") as fp:
        json.dump(config, fp, indent=4)
    return True


def load_session(repo: Union[Path, None] = None) -> bool:
    """Load rules, tokens and config from a repository, and create
    Python objects in memory to work with them.

    Args:
        repo (Path, optional): Absolute path to a repository. Defaults to None.

    Returns:
        bool: True if loading session operation was successful.
    """
    repo: Path = repo or get_repo()
    if not repo.exists():
        logger.warning(f"Given repo directory does not exist: {repo}")
        return False
    namingconf = repo / "vfxnaming.conf"
    if not namingconf.exists():
        logger.warning(f"Repo is not valid. vfxnaming.conf not found {namingconf}")
        return False
    rules.reset_rules()
    tokens.reset_tokens()
    # tokens and rules
    for dirpath, dirnames, filenames in os.walk(repo):
        for filename in filenames:
            filepath = Path(dirpath) / filename
            if filename.endswith(".token"):
                logger.debug(f"Loading token: {filepath}")
                tokens.load_token(filepath)
            elif filename.endswith(".rule"):
                logger.debug(f"Loading rule: {filepath}")
                rules.load_rule(filepath)
    # extra configuration
    if namingconf.exists():
        logger.debug(f"Loading active rule: {namingconf}")
        with open(namingconf) as fp:
            config = json.load(fp)
        rules.set_active_rule(config.get("set_active_rule"))
    return True



from typing import Dict
import copy


class Serializable(object):
    def data(self) -> Dict:
        """Collect all data for this object instance.

        Returns:
            dict: {attribute:value}
        """
        retval = copy.deepcopy(self.__dict__)
        retval["_Serializable_classname"] = type(self).__name__
        retval["_Serializable_version"] = "1.0"
        return retval

    @classmethod
    def from_data(cls, data: Dict) -> "Serializable":
        """Create object instance from give data. Used by Rule,
        Token, Separator to create object instances from disk saved data.

        Args:
            data (dict): {attribute:value}

        Returns:
            Serializable: Object instance for Rule, Token or Separator.
        """
        # Validation
        if data.get("_Serializable_classname") != cls.__name__:
            return None
        del data["_Serializable_classname"]
        if data.get("_Serializable_version") is not None:
            del data["_Serializable_version"]

        this = cls(None)
        this.__dict__.update(data)
        return this



from vfxnaming.naming import parse, solve, validate, get_repo, save_session, load_session  # noqa: F401
from vfxnaming.rules import (  # noqa: F401
    add_rule,
    remove_rule,
    has_rule,
    reset_rules,
    get_active_rule,
    set_active_rule,
    get_rule,
    get_rules,
    save_rule,
    load_rule,
    Rule,
)
from vfxnaming.tokens import (  # noqa: F401
    add_token,
    add_token_number,
    remove_token,
    has_token,
    reset_tokens,
    get_token,
    get_tokens,
    save_token,
    load_token,
    Token,
    TokenNumber,
)
from vfxnaming.error import ParsingError, SolvingError, TokenError  # noqa: F401
